{-# LANGUAGE OverloadedStrings #-}

module Main where

import Prelude hiding (catch)

import           Control.Applicative
import           Control.Exception
import           Control.Monad
import           Control.Monad.IO.Class
import qualified Crypto.Hash.MD5 as Crypto
import           Data.ByteString (ByteString)
import qualified Data.ByteString as B
import qualified Data.ByteString.Lazy as LB
import qualified Data.ByteString.Char8 as BC
import qualified Data.ByteString.Base64 as B64
import           Data.Default (def)
import qualified Data.Text as T
import qualified Language.Fay as Fay
import           Snap.Core
import           Snap.Http.Server (httpServe, setPort, defaultConfig)
import           Snap.Util.FileServe (serveDirectory, serveFileAs)
import           System.Directory
import           System.FilePath

main :: IO ()
main = httpServe (setPort 8000 defaultConfig) $
    path "compile" compileJS <|> serveDirectory "static"

compileJS :: Snap ()
compileJS = do
    body <- LB.toStrict <$> readRequestBody maxSourceSize
    let hash = hashedSource body
    let source = "user" </> hash ++ ".hs"
    let target = "user" </> hash ++ ".js"
    let errors = "user" </> hash ++ ".err"
    whenIO (not <$> eitherExists target errors) $ liftIO $ do
        B.writeFile source body
        compileUserSource source target errors
    whenIO (doesFileExist target) $ do
        serveFileAs "text/javascript" target
    whenIO (doesFileExist errors) $ do
        modifyResponse (setResponseCode 500)
        modifyResponse (setContentType "text/plain")
        sendFile errors
  where maxSourceSize = 2000000
        eitherExists a b = (||) <$> doesFileExist a <*> doesFileExist b

whenIO :: MonadIO m => IO Bool -> m () -> m ()
whenIO p x = do cond <- liftIO p ; when cond x

hashedSource :: ByteString -> FilePath
hashedSource = map slashToDash . BC.unpack . B64.encode . Crypto.hash
  where slashToDash '/' = '-'
        slashToDash c   = c

compileUserSource :: FilePath -> FilePath -> FilePath -> IO ()
compileUserSource source target errors = do
    result <- safeCompile source
    case result of
        Left err -> writeFile errors err
        Right js -> writeFile target js

safeCompile :: FilePath -> IO (Either String String)
safeCompile source = left (sanitizeError source) <$> compile `catch` handleError
  where compile = left Fay.showCompileError <$>
                       Fay.compileFile (compileConfig source) source
        handleError (ErrorCall str) = return (Left str)

sanitizeError :: String -> String -> String
sanitizeError source = T.unpack
                     . T.replace (T.pack source) "your program"
                     . T.pack

compileConfig :: FilePath -> Fay.CompileConfig
compileConfig fp = def {
    Fay.configFilePath = Just fp,
    Fay.configOptimize = True,
    Fay.configTypecheck = True,
    Fay.configDirectoryIncludes = [ "client" ],
    Fay.configPrettyPrint = True
    }

left :: (a -> b) -> Either a c -> Either b c
left f (Left x)  = Left (f x)
left _ (Right x) = Right x

