module CodeWorld (
    -- Re-exports from the prelude until Fay removes the need to import it
    -- explicitly.  Saves a line of code per program.
    Fay,
    Char,
    String,
    Integer,
    Double,
    Int,
    Bool(..),
    Show,
    Read,
    Maybe(..),
    Typeable(..),
    Data(..),
    Monad,
    Eq(..),
    read,
    (>>),
    (>>=),
    (+),
    (-),
    (*),
    (/),
    (>),
    (<),
    (>=),
    (<=),
    (||),
    (&&),
    fail,
    return,
    force,
    ($),
    ($!),
    (!!),
    (++),
    (.),
    (=<<),
    (**),
    (^^),
    (^),
    Defined(..),
    Either(..),
    Ordering(..),
    abs,
    acos,
    acosh,
    all,
    and,
    any,
    asin,
    asinh,
    asTypeOf,
    atan,
    atanh,
    break,
    ceiling,
    compare,
    concat,
    concatMap,
    const,
    cos,
    cosh,
    curry,
    cycle,
    div,
    divMod,
    drop,
    dropWhile,
    either,
    elem,
    enumFrom,
    enumFromThen,
    enumFromThenTo,
    enumFromTo,
    error,
    even,
    exp,
    filter,
    find,
    flip,
    floor,
    foldl,
    foldl1,
    foldr,
    foldr1,
    forM_,
    fromInteger,
    fromIntegral,
    fromRational,
    fst,
    gcd,
    head,
    id,
    init,
    insertBy,
    intercalate,
    intersperse,
    iterate,
    last,
    lcm,
    length,
    lines,
    log,
    logBase,
    lookup,
    map,
    mapM_,
    max,
    maximum,
    maybe,
    min,
    minimum,
    mod,
    negate,
    not,
    notElem,
    nub,
    null,
    odd,
    or,
    otherwise,
    pi,
    pred,
    prependToAll,
    print,
    product,
    properFraction,
    putStrLn,
    quot,
    quotRem,
    recip,
    rem,
    repeat,
    replicate,
    reverse,
    round,
    scanl,
    scanl1,
    scanr,
    scanr1,
    seq,
    sequence,
    sequence_,
    show,
    signum,
    sin,
    sinh,
    snd,
    sort,
    sortBy,
    span,
    splitAt,
    sqrt,
    subtract,
    succ,
    sum,
    tail,
    take,
    takeWhile,
    tan,
    tanh,
    truncate,
    uncurry,
    undefined,
    unlines,
    until,
    unwords,
    unzip,
    unzip3,
    when,
    words,
    zip,
    zip3,
    zipWith,
    zipWith3,

    -- Color
    Color(..),
    black,
    white,
    red,
    green,
    blue,
    cyan,
    magenta,
    yellow,
    aquamarine,
    orange,
    azure,
    violet,
    chartreuse,
    rose,
    brown,
    pink,
    purple,
    gray,
    grey,
    light,
    dark,
    translucent,

    -- Picture
    Point,
    Vector,
    Picture,
    blank,
    line,
    polygon,
    thickLine,
    rectangle,
    solidRectangle,
    thickRectangle,
    circle,
    solidCircle,
    thickCircle,
    arc,
    sector,
    thickArc,
    text,
    color,
    translate,
    scale,
    rotate,
    pictures,

    -- Random
    StdGen,
    newStdGen,
    randomR,
    splitR,

    -- Event
    GameEvent(..),

    -- Entry points
    displayInCanvas,
    animateInCanvas,
    simulateInCanvas,
    playInCanvas
    ) where

import Language.Fay.Prelude
import Language.Fay.FFI
import Web
import Color
import Picture
import Random
import Event

-------------------------------------------------------------------------
-- Bindings to special functions defined in codeworld.js

data Event
instance Foreign Event

getSpecialKey :: Event -> Fay String
getSpecialKey = ffi "window.getSpecialKey(%1)"

getPressedKey :: Event -> Fay String
getPressedKey = ffi "window.getPressedKey(%1)"

getReleasedKey :: Event -> Fay String
getReleasedKey = ffi "window.getReleasedKey(%1)"

stopEvent :: Event -> Fay ()
stopEvent = ffi "window.stopEvent(%1)"

mouseToElementX :: Event -> Fay Double
mouseToElementX = ffi "window.mouseToElementX(%1)"

mouseToElementY :: Event -> Fay Double
mouseToElementY = ffi "window.mouseToElementY(%1)"

getEventMouseButton :: Event -> Fay Int
getEventMouseButton = ffi "%1['button']"

setKeyDownEvent :: (Event -> Fay Bool) -> Fay ()
setKeyDownEvent = ffi "window.codeworldKeyDownEvent=%1"

setKeyUpEvent :: (Event -> Fay Bool) -> Fay ()
setKeyUpEvent = ffi "window.codeworldKeyUpEvent=%1"

setKeyPressEvent :: (Event -> Fay Bool) -> Fay ()
setKeyPressEvent = ffi "window.codeworldKeyPressEvent=%1"

setMouseDownEvent :: (Event -> Fay Bool) -> Fay ()
setMouseDownEvent = ffi "window.codeworldMouseDownEvent=%1"

setMouseUpEvent :: (Event -> Fay Bool) -> Fay ()
setMouseUpEvent = ffi "window.codeworldMouseUpEvent=%1"

setMouseMoveEvent :: (Event -> Fay Bool) -> Fay ()
setMouseMoveEvent = ffi "window.codeworldMouseMoveEvent=%1"

setInterval :: Double -> Fay () -> Fay ()
setInterval = ffi "window.codeworldSetInterval(%1,%2)"

clearInterval :: Fay ()
clearInterval = ffi "window.codeworldClearInterval()"

cleanup :: Fay ()
cleanup = do
    setKeyDownEvent $ const $ return True
    setKeyUpEvent $ const $ return True
    setKeyPressEvent $ const $ return True
    setMouseDownEvent $ const $ return True
    setMouseUpEvent $ const $ return True
    setMouseMoveEvent $ const $ return True
    clearInterval

-------------------------------------------------------------------------
-- Draw state.  The draw state consists of the current affine
-- transformation matrix to apply when drawing, and whether a color has
-- been fixed yet (the latter ensures that applying color to a picture
-- that previously had colors in it will work)

type DrawState = (Double, Double, Double, Double, Double, Double, Bool)

translateDrawState :: Double -> Double -> DrawState -> DrawState
translateDrawState x y (a,b,c,d,e,f,hc) =
    (a, b, c, d, a * x + c * y + e, b * x + d * y + f, hc)

scaleDrawState :: Double -> Double -> DrawState -> DrawState
scaleDrawState x y (a,b,c,d,e,f,hc) =
    (x*a, x*b, y*c, y*d, e, f, hc)

rotateDrawState :: Double -> DrawState -> DrawState
rotateDrawState r (a,b,c,d,e,f,hc) = let th = r * pi / 180 in
    (a * cos th + c * sin th,
     b * cos th + d * sin th,
     c * cos th - a * sin th,
     d * cos th - b * sin th,
     e, f, hc)

hasColor :: DrawState -> Bool
hasColor (_,_,_,_,_,_,hc) = hc

setColor :: DrawState -> DrawState
setColor (a,b,c,d,e,f,_) = (a,b,c,d,e,f,True)

withTransform :: Context -> DrawState -> Fay () -> Fay ()
withTransform ctx (a,b,c,d,e,f,_) action = do
    save ctx
    transform ctx a b c d e f
    beginPath ctx
    action
    restore ctx

-- Follow a list of points to build a path in the given Context
pathFromPoints :: Context -> [Point] -> Fay ()
pathFromPoints _   [] = return ()
pathFromPoints ctx ((sx,sy):ps) = do
    moveTo ctx sx sy
    forM_ ps $ \(x,y) -> lineTo ctx x y

drawPicture :: Context -> DrawState -> Picture -> Fay ()
drawPicture ctx t (Polygon ps) = do
    withTransform ctx t $ pathFromPoints ctx ps
    fill ctx

drawPicture ctx t (Line ps w) = do
    save ctx
    withTransform ctx t $ pathFromPoints ctx ps
    when (w > 0) $ setLineWidth ctx w
    stroke ctx
    restore ctx

drawPicture ctx t (Arc b e r w) = do
    save ctx
    withTransform ctx t $ do
        when (r > 0) $ canvasArc ctx 0 0 r (b*pi/180) (e*pi/180) False
    when (w > 0) $ setLineWidth ctx w
    stroke ctx
    restore ctx

drawPicture ctx t (Text txt) =
    withTransform ctx t $ do
        canvasScale ctx 1 (0-1)
        fillText ctx txt 0 0

drawPicture ctx t (Color (RGBA r g b a) p)
  | hasColor t = drawPicture ctx t p
  | otherwise = do
      let str = "rgba(" ++ show (r * 100) ++ "%,"
                        ++ show (g * 100) ++ "%,"
                        ++ show (b * 100) ++ "%,"
                        ++ show a ++ ")"
      save ctx
      setStrokeStyle ctx str
      setFillStyle ctx str
      drawPicture ctx (setColor t) p
      restore ctx

drawPicture ctx t (Translate x y p) = drawPicture ctx (translateDrawState x y t) p
drawPicture ctx t (Scale x y p)     = drawPicture ctx (scaleDrawState x y t) p
drawPicture ctx t (Rotate r p)      = drawPicture ctx (rotateDrawState r t) p
drawPicture ctx t (Pictures ps)     = mapM_ (drawPicture ctx t) (reverse ps)

withCanvas :: (Element -> Context -> Fay ()) -> Fay ()
withCanvas go = do
    canvas <- getElementById "canvas"
    ctx    <- getContext canvas "2d"
    go canvas ctx

drawOn :: Context -> Picture -> Fay ()
drawOn ctx pic = do
    clearRect ctx 0 0 500 500
    save ctx
    canvasTranslate ctx 250 250
    canvasScale ctx 1 (0-1)
    setTextAlign ctx "left"
    setTextBaseline ctx "alphabetic"
    setLineWidth ctx 0
    setFont ctx "16px Times Roman"
    drawPicture ctx (1,0,0,1,0,0,False) pic
    restore ctx

displayInCanvas :: Picture -> Fay ()
displayInCanvas pic = withCanvas $ \ canvas ctx -> do
    cleanup
    drawOn ctx pic

animateInCanvas :: (Double -> Picture) -> Fay ()
animateInCanvas anim = withCanvas $ \ canvas ctx -> do
    cleanup
    startTime <- currentTimeMillis
    setInterval 30 $ do
        currentTime <- currentTimeMillis
        let t = (currentTime - startTime) / 1000
        drawOn ctx (anim t)
    return ()

data SimState a = SimState a
instance Foreign (SimState a)

withSimState :: (a -> a) -> Ref (SimState a) -> Fay a
withSimState f ref = do
    SimState val <- readRef ref
    let newVal = f val
    writeRef ref (SimState newVal)
    return newVal

simulateInCanvas :: (StdGen -> a)
                 -> (Double -> a -> a)
                 -> (a -> Picture)
                 -> Fay ()
simulateInCanvas i s d = withCanvas $ \ canvas ctx -> do
    cleanup
    startTime <- currentTimeMillis
    g <- newStdGen
    valueRef <- newRef (SimState (i g))
    timeRef <- newRef startTime
    setInterval 30 $ do
        lastTime     <- readRef timeRef
        currentTime  <- currentTimeMillis
        let dt     = (currentTime - lastTime) / 1000
        val <- withSimState (s dt) valueRef
        writeRef timeRef currentTime
        drawOn ctx (d val)
    return ()

playInCanvas :: (StdGen -> a)
             -> (Double -> a -> a)
             -> (GameEvent -> a -> a)
             -> (a -> Picture)
             -> Fay ()
playInCanvas i s e d = withCanvas $ \ canvas ctx -> do
    cleanup
    startTime <- currentTimeMillis
    g <- newStdGen
    valueRef <- newRef (SimState (i g))
    timeRef <- newRef startTime
    setKeyDownEvent $ \ev -> do
        k <- getSpecialKey ev
        if k == "None" then return True else do
            withSimState (e (KeyPressEvent k)) valueRef
            return True
    setKeyPressEvent $ \ev -> do
        k <- getPressedKey ev
        if k == "None" then return True else do
            withSimState (e (KeyPressEvent k)) valueRef
            return True
    setKeyUpEvent $ \ev -> do
        k <- getReleasedKey ev
        if k == "None" then return True else do
            withSimState (e (KeyReleaseEvent k)) valueRef
            return True
    setMouseDownEvent $ \ev -> do
        focusElement canvas
        x <- mouseToElementX ev
        y <- mouseToElementY ev
        when (abs x < 250 && abs y < 250) $ do
            b <- getEventMouseButton ev
            withSimState (e (MousePressEvent b (x,y))) valueRef
        return True
    setMouseUpEvent $ \ev -> do
        x <- mouseToElementX ev
        y <- mouseToElementY ev
        when (abs x < 250 && abs y < 250) $ do
            b <- getEventMouseButton ev
            withSimState (e (MouseReleaseEvent b (x,y))) valueRef
        return True
    setMouseMoveEvent $ \ev -> do
        x <- mouseToElementX ev
        y <- mouseToElementY ev
        when (abs x < 250 && abs y < 250) $ do
            withSimState (e (MouseMoveEvent (x,y))) valueRef
        return True
    setInterval 30 $ do
        lastTime     <- readRef timeRef
        currentTime  <- currentTimeMillis
        let dt     = (currentTime - lastTime) / 1000
        val <- withSimState (s dt) valueRef
        writeRef timeRef currentTime
        drawOn ctx (d val)
    return ()

