module Picture where

import Language.Fay.Prelude
import Color

type Point = (Double, Double)
type Vector = Point

data Picture = Polygon [Point]
             | Line [Point] Double
             | Arc Double Double Double Double
             | Text String
             | Color Color Picture
             | Translate Double Double Picture
             | Scale Double Double Picture
             | Rotate Double Picture
             | Pictures [Picture]

-- A blank picture
blank :: Picture
blank = Pictures []

-- A thin line with these points as endpoints
line :: [Point] -> Picture
line ps = Line ps 0

-- A solid polygon with these points as vertices
polygon :: [Point] -> Picture
polygon = Polygon

-- A thick line, with these endpoints, with this line width
thickLine :: [Point] -> Double -> Picture
thickLine = Line

-- A thin rectangle, with this width and height
rectangle :: Double -> Double -> Picture
rectangle w h = line [
    (0-w/2, 0-h/2), (w/2, 0-h/2), (w/2, h/2), (0-w/2, h/2), (0-w/2, 0-h/2)
    ]

-- A solid rectangle, with this width and height
solidRectangle :: Double -> Double -> Picture
solidRectangle w h = polygon [
    (0-w/2, 0-h/2), (w/2, 0-h/2), (w/2, h/2), (0-w/2, h/2)
    ]

-- A thick rectangle, with this width and height and line width
thickRectangle :: Double -> Double -> Double -> Picture
thickRectangle w h lw = thickLine [
    (0-w/2, 0-h/2), (w/2, 0-h/2), (w/2, h/2), (0-w/2, h/2), (0-w/2, 0-h/2)
    ] lw

-- A thin circle, with this radius
circle :: Double -> Picture
circle = arc 0 360

-- A solid circle, with this radius
solidCircle :: Double -> Picture
solidCircle r = thickCircle (r/2) r

-- A thick circle, with this radius and line width
thickCircle :: Double -> Double -> Picture
thickCircle = Arc 0 360

-- A thin arc, starting and ending at these angles, with this radius
arc :: Double -> Double -> Double -> Picture
arc b e r = Arc b e r 0

-- A solid sector of a circle (i.e., a pie slice) starting and ending at these
-- angles, with this radius
sector :: Double -> Double -> Double -> Picture
sector b e r = Arc b e (r/2) r

-- A thick arc, starting and ending at these angles, with this radius and
-- line width
thickArc :: Double -> Double -> Double -> Double -> Picture
thickArc = Arc

-- A piece of text
text :: String -> Picture
text = Text

-- A picture drawn entirely in this color.
color :: Color -> Picture -> Picture
color = Color

-- A picture drawn translated in these directions.
translate :: Double -> Double -> Picture -> Picture
translate = Translate

-- A picture scaled by these factors.
scale :: Double -> Double -> Picture -> Picture
scale = Scale

-- A picture scaled by this angle.
rotate :: Double -> Picture -> Picture
rotate = Rotate

-- A picture made by drawing these pictures, ordered from top to bottom.
pictures :: [Picture] -> Picture
pictures = Pictures

